/**************************************************************************/
/*                                                                        */
/*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       */
/*  Kim Nguyen, Laurent Sartre                                            */
/*                                                                        */
/*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      */
/*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 */
/*                                                                        */
/*  https://www.informatique-mpi.fr/                                      */
/*                                                                        */
/**************************************************************************/

// Recherche d'un motif dans un texte avec l'algorithme de Rabin--Karp

#include "prelude.h"
#include "arith.h"

// base du calcul
const uint64_t B = 256;
// un nombre premier tel que P*P tient dans uint64_t
const uint64_t P = 1869461003;

// le hachage des `len` premiers caractères de `s`
uint64_t rk_hash(char *s, int len) {
  uint64_t h = 0;
  for (int i = 0; i < len; i++) {
    unsigned char c = s[i];
    h = (h * B + c) % P;
  }
  return h;
}

// affiche toutes les occurrences de `m` dans `t` avec Rabin--Karp
// et renvoie le nombre d'occurrences

// note : si le motif `m` est vide, renvoie directement `strlen(t)+1`
// (sans afficher les chaînes "occurrence ...")
int rabin_karp(char *m, char *t) {
  int lm = strlen(m), lt = strlen(t);
  if (lm == 0) return lt+1;
  if (lt < lm) return 0;
  uint64_t sh = arith_power_mod(B, lm - 1, P);
  uint64_t hm = rk_hash(m, lm), ht = rk_hash(t, lm);
  int count = 0;
  for (int i = 0; true; i++) {
    if (ht == hm && strncmp(t + i, m, lm) == 0) {
      printf("occurrence à la position %d\n", i);
      count++;
    }
    if (i == lt - lm) break;
    unsigned char ci = t[i]; // on enlève le caractère t[i]
    ht = (ht + P - (ci * sh) % P) % P;
    unsigned char c = t[i + lm]; // et on ajoute le caractère t[i+lm]
    ht = (ht * B + c) % P;
  }
  return count;
}
